/*
 * neuron.c
 *
 * Created: 05/22/2012 8:41:42 PM
 * Author:		Craig Berscheidt (craig.berscheidt@built-to-spec.com)
 * Compiler:	gcc-avr using AVR Studio 5
 * Target:		ATTINY45
 *
 * Distributed under Creative Commons 3.0 -- Attib & Share Alike
 *
 * Program Command (avrdude using the USBtinyISP):
 * avrdude -c usbtiny -p t85 -U flash:w:"neuron.hex"
 *
 * Peripherals Used:
 * Timer 0 - PWM for tone generation
 * Timer 1 - Overflow interrupt generation for sampling the LEDs
 */

#define F_CPU 1000000UL

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <util/delay.h>

#define NUM_SAMPLES 4		// Number of samples to average for detection algorithm 
#define PROP_DELAY_MS 350	// Time to delay after light detected before blinking in neuron mode (ms)
#define PULSE_TIME_MS 350	// Time for LED to be on when in neuron mode (ms)
#define BLINK_DELAY 200		// Time between LED toggling in blink mode (ms)

#define SAMPLE_PRESCALE		(1<<CS13)	// 1/128th the system clock, works out to ~30.5Hz sample rate (timer 1)
//Calculate out the number of sample cycles based on the specified prescaler
#define PROP_DELAY ((PROP_DELAY_MS * ((F_CPU / (1 << (SAMPLE_PRESCALE - 1))) / 256 ))/1000)
#define PULSE_TIME ((PULSE_TIME_MS * ((F_CPU / (1 << (SAMPLE_PRESCALE - 1))) / 256 ))/1000)

#define CONV_TIMEOUT 150

#define RANDOM_TONE	1 		// If RANDOM_TONE == 1 a random tone will be played each activation in neuron mode  
#define TONE_VAL	0x80	// TONE_VAL corresponds to the value in OCR0A needed to generate a specific tone

//TODO - Add Tone Chart

// Init Defs
#define DDRB_INIT 0xFF		// All outputs
#define PORTB_INIT 0x00

// Define the board version here, corresponds to different pin mappings defined below
// Version 0 boards don't have any version info, version 1 boards are labeled "Neuron v1.1"
#define BOARD_VER 0

enum pin_dirs {
	INPUT = 0,
	OUTPUT = 1
};

enum LED_states {
	LED_INIT,
	LED_SAMPLING,
	LED_PROPAGATING,
	LED_BLINKING
};

typedef struct {
	uint8_t p_pin;
	uint8_t n_pin;
	uint8_t adc_channel;
	uint8_t state;  //TODO - Typedef this
	int count;
} led;

// Pin Definitions
#define JUMPER_PIN PB0

#if BOARD_VER == 1
	led leds[] = {
		{
			.p_pin = 1,
			.n_pin = 4,
			.adc_channel = 2,
			.state = LED_INIT,
			.count = 0
		},
		{
			.p_pin = 3,
			.n_pin = 2,
			.adc_channel = 1,
			.state = LED_INIT,
			.count = 0
		}	
	};
#else
	led leds[] = {
		{
			.p_pin = 1,
			.n_pin = 2,
			.adc_channel = 1,
			.state = LED_INIT,
			.count = 0
		},
		{
			.p_pin = 3,
			.n_pin = 4,
			.adc_channel = 2,
			.state = LED_INIT,
			.count = 0
		}	
	};
#endif

void set_gpio(int gpio, int value)
{
    if(value == 0)
		PORTB &= ~(_BV(gpio));
    else
		PORTB |= _BV(gpio);
}

int read_gpio(int gpio)
{
	if (PINB & _BV(gpio))
		return 1;
	else
		return 0;
}

void set_dir(int gpio, int dir)
{
	if (dir == INPUT)
		DDRB &= ~(_BV(gpio));
	else
		DDRB |= _BV(gpio);
}

int get_light(int p_pin, int n_pin, int timeout)
{
	int i;
	
	// Apply reverse voltage
	set_dir(p_pin, OUTPUT);
	set_dir(n_pin, OUTPUT);
	set_gpio(p_pin, 0);
	set_gpio(n_pin, 1);
	
	// Isolate the N side
	set_dir(n_pin, INPUT);
	set_gpio(n_pin, 0);
	
	// Count how long it takes the diode to bleed back down to a logic zero
	for ( i = 0; i < timeout; i++)
		if (read_gpio(n_pin) == 0) break;
	
	return i;
}

void enable_tone(int enable)
{
	static int enabled = 0;
	
	// Start or stop tone generation
	if ((enabled == 0) && (enable == 1))
	{
		set_dir(JUMPER_PIN, OUTPUT);
		
		TCCR0A |= _BV(COM0A0)	//Toggle OC2A on match
				| _BV(WGM01);	//CTC Mode
		
		TCCR0B |= 0x3;			//prescaler = 64
		TCNT0 = 0x00;			//Clear count
		//Set OCR0A to adjust frequency of output
#if RANDOM_TONE == 1
		OCR0A = rand() % 100;
#else
		OCR0A = TONE_VAL;
#endif
		
		enabled = 1;
	}
	else if ((enabled == 1) && (enable == 0))
	{
		set_dir(JUMPER_PIN, INPUT);
		TCCR0A = 0x00;		//Normal operation
		TCCR0B = 0x00;		//Normal operation
		TCNT0 = 0x00;		//Clear count
		enabled = 0;
	}
}

void blink(void)
{
	// Set up pins - TODO: Clean this up a bit
	set_dir(leds[0].p_pin, OUTPUT);
	set_dir(leds[1].p_pin, OUTPUT);
	set_dir(leds[0].n_pin, OUTPUT);
	set_dir(leds[1].n_pin, OUTPUT);
	set_gpio(leds[0].n_pin, 0);
	set_gpio(leds[1].n_pin, 0);
	set_dir(JUMPER_PIN, INPUT);
	set_gpio(JUMPER_PIN, 1);
	
	// Stay in blink mode until jumper is removed
	while(read_gpio(JUMPER_PIN) == 0)
	{
		// Turn on one LED, then the other
		set_gpio(leds[0].p_pin, 0);
		set_gpio(leds[1].p_pin, 1);
		_delay_ms(BLINK_DELAY);
		set_gpio(leds[0].p_pin, 1);
		set_gpio(leds[1].p_pin, 0);
		_delay_ms(BLINK_DELAY);	
	}
	
	// Turn off LEDs again
	set_gpio(leds[0].p_pin, 0);
}

ISR(TIM1_OVF_vect)
{
	static int samples[2][NUM_SAMPLES];
	static int i;
	int j, k, avg;
	
	// Check the state and take appropriate actions for both LEDs
	for(j = 0; j < 2; j++)
	{
		_delay_us(100); // TODO - find out why this delay is needed

		// if the LED is not actively on, take a sample
		if( leds[j].state != LED_BLINKING )
		{
			//Take sample
			samples[j][i] = get_light(leds[j].p_pin, leds[j].n_pin, CONV_TIMEOUT);
				
			// If the LED is not taking initial samples, average the
			// samples and determine if we need to trigger something
			if( leds[j].state != LED_INIT )
			{
				avg = 0;
				for(k = 0; k < NUM_SAMPLES; k++)
					avg += samples[j][k];
						
				avg /= NUM_SAMPLES;
					
				if( ( (avg/2) > samples[j][i] ) && ( leds[(!j)].state == LED_SAMPLING ) )  // TODO - move sampling check up one level?
				{
					// Put opposite LED into the propagating state if it's idle and set the count
					leds[(!j)].count = 0;
					leds[(!j)].state = LED_PROPAGATING;
				}

				// If we're waiting to blink our LED, update count and see if we're ready
				if( leds[j].state == LED_PROPAGATING )
				{
					leds[j].count++;
					if( leds[j].count >= PROP_DELAY )
					{
						// We need to turn on the LED (and sound) and move to the blinking state
						leds[j].count = 0;
						leds[j].state = LED_BLINKING;
						enable_tone(1);
						// Turn on LED
						set_dir(leds[j].p_pin, OUTPUT);
						set_dir(leds[j].n_pin, OUTPUT);
						set_gpio(leds[j].p_pin, 1);
						set_gpio(leds[j].n_pin, 0);
					}
				}
			}
			else
			{
				// We're currently in the initialization state, see if we've finished gathering samples
				leds[j].count++;
				if ( leds[j].count >= NUM_SAMPLES )
				{
					// If we've got or initial samples, go to the polling state
					leds[j].count = 0;
					leds[j].state = LED_SAMPLING;
				}
			}
		}
		else
		{
			// We've currently got the LED on, we need to update our count and see if we're done blinking
			leds[j].count++;
			if ( leds[j].count >= PULSE_TIME )
			{
				// We're done blinking, shut off the LED (and sound) and return to the initialization state
				// to get a new baseline for ambient lighting conditions (so we don't blink again on accident)
				leds[j].state = LED_INIT;
				enable_tone(0);
				set_dir(leds[j].p_pin, OUTPUT);
				set_dir(leds[j].n_pin, OUTPUT);
				set_gpio(leds[j].p_pin, 0);      //TODO - possibly redundant
				set_gpio(leds[j].n_pin, 0);
			}
		}	
	}
		
	// Update i, which is used for the current sample #
	i++;
	i = i % NUM_SAMPLES;
}

int main(void)
{	
	//Port Initialization
	DDRB = DDRB_INIT;
	PORTB = PORTB_INIT;
	
	//Timer Initialization
	TCCR1 |= SAMPLE_PRESCALE;	// prescale timer 1
	TIMSK |=1<<TOIE1;			// enable timer overflow interrupt
	sei();						// enable interrupts
	
    while(1)
    {
		cli();
		// Check the jumper state if we're not generating a tone to see if we need to go into blinky mode
		if ( (leds[0].state != LED_BLINKING) && (leds[1].state != LED_BLINKING) )
		{
			// Set up jumper pin as input (PB0) and enable pull up, then read its state to determine if jumper present
			set_dir(JUMPER_PIN, INPUT);
			set_gpio(JUMPER_PIN, 1);
			
			// Go into blink mode
			if(read_gpio(JUMPER_PIN) == 0)
			{
				blink();
				leds[0].state = LED_INIT;
				leds[1].state = LED_INIT;
			}		
		}
		sei();
		_delay_ms(250);				
	}
}